@@ -2,10 +2,27 @@
 Encoding and decoding python native data structures as
 portable JData-spec annotated dict structure
 
-Copyright (c) 2019-2022 Qianqian Fang <q.fang at neu.edu>
+Copyright (c) 2019-2025 Qianqian Fang <q.fang at neu.edu>
 """
 
-__all__ = ["encode", "decode", "jdtype", "jsonfilter"]
+__all__ = [
+    "encode",
+    "decode",
+    "jdataencode",
+    "jdatadecode",
+    "jdtype",
+    "jsonfilter",
+    "zlibencode",
+    "zlibdecode",
+    "gzipencode",
+    "gzipdecode",
+    "lzmaencode",
+    "lzmadecode",
+    "lz4encode",
+    "lz4decode",
+    "base64encode",
+    "base64decode",
+]
 
 ##====================================================================================
 ## dependent libraries
@@ -15,6 +32,8 @@ import numpy as np
 import copy
 import zlib
 import base64
+import os
+import re
 
 ##====================================================================================
 ## global variables
@@ -64,7 +83,7 @@ _allownumpy = ("_ArraySize_", "_ArrayData_", "_ArrayZipSize_", "_ArrayZipData_")
 ##====================================================================================
 
 
-def encode(d, opt={}):
+def encode(d, opt={}, **kwargs):
     """@brief Encoding a Python data structure to portable JData-annotated dict constructs
 
     This function converts complex data types (usually not JSON-serializable) into
@@ -79,6 +98,7 @@ def encode(d, opt={}):
     """
 
     opt.setdefault("inplace", False)
+    opt.update(kwargs)
 
     if "compression" in opt:
         if opt["compression"] == "lzma":
@@ -131,12 +151,19 @@ def encode(d, opt={}):
         return newobj
     elif isinstance(d, np.ndarray) or np.iscomplex(d):
         newobj = {}
-        newobj["_ArrayType_"] = jdtype[str(d.dtype)] if (str(d.dtype) in jdtype) else str(d.dtype)
+        newobj["_ArrayType_"] = (
+            jdtype[str(d.dtype)] if (str(d.dtype) in jdtype) else str(d.dtype)
+        )
         if np.isscalar(d):
             newobj["_ArraySize_"] = 1
         else:
             newobj["_ArraySize_"] = list(d.shape)
-        if d.dtype == np.complex64 or d.dtype == np.complex128 or d.dtype == np.csingle or d.dtype == np.cdouble:
+        if (
+            d.dtype == np.complex64
+            or d.dtype == np.complex128
+            or d.dtype == np.csingle
+            or d.dtype == np.cdouble
+        ):
             newobj["_ArrayIsComplex_"] = True
             newobj["_ArrayData_"] = np.stack((d.ravel().real, d.ravel().imag))
         else:
@@ -156,17 +183,26 @@ def encode(d, opt={}):
             elif opt["compression"] == "gzip":
                 gzipper = zlib.compressobj(wbits=(zlib.MAX_WBITS | 16))
                 newobj["_ArrayZipData_"] = gzipper.compress(newobj["_ArrayZipData_"])
+                newobj["_ArrayZipData_"] += gzipper.flush()
             elif opt["compression"] == "lzma":
                 try:
-                    newobj["_ArrayZipData_"] = lzma.compress(newobj["_ArrayZipData_"], lzma.FORMAT_ALONE)
+                    newobj["_ArrayZipData_"] = lzma.compress(
+                        newobj["_ArrayZipData_"], lzma.FORMAT_ALONE
+                    )
                 except Exception:
-                    print('you must install "lzma" module to compress with this format, ignoring')
+                    print(
+                        'you must install "lzma" module to compress with this format, ignoring'
+                    )
                     pass
             elif opt["compression"] == "lz4":
                 try:
-                    newobj["_ArrayZipData_"] = lz4.frame.compress(newobj["_ArrayZipData_"].tobytes())
+                    newobj["_ArrayZipData_"] = lz4.frame.compress(
+                        newobj["_ArrayZipData_"].tobytes()
+                    )
                 except ImportError:
-                    print('you must install "lz4" module to compress with this format, ignoring')
+                    print(
+                        'you must install "lz4" module to compress with this format, ignoring'
+                    )
                     pass
             elif opt["compression"].startswith("blosc2"):
                 try:
@@ -182,14 +218,18 @@ def encode(d, opt={}):
                         blosc2nthread = opt["nthread"]
                     newobj["_ArrayZipData_"] = blosc2.compress2(
                         newobj["_ArrayZipData_"],
-                        compcode=BLOSC2CODEC[opt["compression"]],
+                        codec=BLOSC2CODEC[opt["compression"]],
                         typesize=d.dtype.itemsize,
                         nthreads=blosc2nthread,
                     )
                 except ImportError:
-                    print('you must install "blosc2" module to compress with this format, ignoring')
+                    print(
+                        'you must install "blosc2" module to compress with this format, ignoring'
+                    )
                     pass
-            if (("base64" in opt) and (opt["base64"])) or opt["compression"] == "base64":
+            if (("base64" in opt) and (opt["base64"])) or opt[
+                "compression"
+            ] == "base64":
                 newobj["_ArrayZipData_"] = base64.b64encode(newobj["_ArrayZipData_"])
             newobj.pop("_ArrayData_")
         return newobj
@@ -202,7 +242,7 @@ def encode(d, opt={}):
 ##====================================================================================
 
 
-def decode(d, opt={}):
+def decode(d, opt={}, **kwargs):
     """@brief Decoding a JData-annotated dict construct into native Python data
 
     This function converts portable JData-annotated dict/list constructs back to native Python
@@ -213,9 +253,18 @@ def decode(d, opt={}):
          'nthread': number of decompression thread of the codec is of the blosc2 class, default is 1
     """
 
-    opt.setdefault("inplace", False)
+    from .jfile import jdlink
 
-    if (isinstance(d, str) or type(d) == "unicode") and len(d) <= 6 and len(d) > 4 and d[-1] == "_":
+    opt.setdefault("inplace", False)
+    opt.setdefault("maxlinklevel", 0)
+    opt.update(kwargs)
+
+    if (
+        (isinstance(d, str) or type(d) == "unicode")
+        and len(d) <= 6
+        and len(d) > 4
+        and d[-1] == "_"
+    ):
         if d == "_NaN_":
             return float("nan")
         elif d == "_Inf_":
@@ -233,12 +282,16 @@ def decode(d, opt={}):
                 d["_ArraySize_"] = np.frombuffer(bytearray(d["_ArraySize_"]))
             if "_ArrayZipData_" in d:
                 newobj = d["_ArrayZipData_"]
-                if (("base64" in opt) and (opt["base64"])) or ("_ArrayZipType_" in d and d["_ArrayZipType_"] == "base64"):
+                if (("base64" in opt) and (opt["base64"])) or (
+                    "_ArrayZipType_" in d and d["_ArrayZipType_"] == "base64"
+                ):
                     newobj = base64.b64decode(newobj)
                 if "_ArrayZipType_" in d and d["_ArrayZipType_"] not in _zipper:
                     raise Exception(
                         "JData",
-                        "compression method {} is not supported".format(d["_ArrayZipType_"]),
+                        "compression method {} is not supported".format(
+                            d["_ArrayZipType_"]
+                        ),
                     )
                 if d["_ArrayZipType_"] == "zlib":
                     newobj = zlib.decompress(bytes(newobj))
@@ -258,7 +311,9 @@ def decode(d, opt={}):
 
                         newobj = lz4.frame.decompress(bytes(newobj))
                     except Exception:
-                        print('Warning: you must install "lz4" module to decompress a data record in this file, ignoring')
+                        print(
+                            'Warning: you must install "lz4" module to decompress a data record in this file, ignoring'
+                        )
                         return copy.deepcopy(d) if opt["inplace"] else d
                 elif d["_ArrayZipType_"].startswith("blosc2"):
                     try:
@@ -267,15 +322,23 @@ def decode(d, opt={}):
                         blosc2nthread = 1
                         if "nthread" in opt:
                             blosc2nthread = opt["nthread"]
-                        newobj = blosc2.decompress2(bytes(newobj), as_bytearray=False, nthreads=blosc2nthread)
+                        newobj = blosc2.decompress2(
+                            bytes(newobj), as_bytearray=False, nthreads=blosc2nthread
+                        )
                     except Exception:
-                        print('Warning: you must install "blosc2" module to decompress a data record in this file, ignoring')
+                        print(
+                            'Warning: you must install "blosc2" module to decompress a data record in this file, ignoring'
+                        )
                         return copy.deepcopy(d) if opt["inplace"] else d
-                newobj = np.frombuffer(bytearray(newobj), dtype=np.dtype(d["_ArrayType_"])).reshape(d["_ArrayZipSize_"])
+                newobj = np.frombuffer(
+                    bytearray(newobj), dtype=np.dtype(d["_ArrayType_"])
+                ).reshape(d["_ArrayZipSize_"])
                 if "_ArrayIsComplex_" in d and newobj.shape[0] == 2:
                     newobj = newobj[0] + 1j * newobj[1]
                 if "_ArrayOrder_" in d and (
-                    d["_ArrayOrder_"].lower() == "c" or d["_ArrayOrder_"].lower() == "col" or d["_ArrayOrder_"].lower() == "column"
+                    d["_ArrayOrder_"].lower() == "c"
+                    or d["_ArrayOrder_"].lower() == "col"
+                    or d["_ArrayOrder_"].lower() == "column"
                 ):
                     newobj = newobj.reshape(d["_ArraySize_"], order="F")
                 else:
@@ -285,17 +348,25 @@ def decode(d, opt={}):
                 return newobj
             elif "_ArrayData_" in d:
                 if isinstance(d["_ArrayData_"], str):
-                    newobj = np.frombuffer(d["_ArrayData_"], dtype=np.dtype(d["_ArrayType_"]))
+                    newobj = np.frombuffer(
+                        d["_ArrayData_"], dtype=np.dtype(d["_ArrayType_"])
+                    )
                 else:
-                    newobj = np.asarray(d["_ArrayData_"], dtype=np.dtype(d["_ArrayType_"]))
+                    newobj = np.asarray(
+                        d["_ArrayData_"], dtype=np.dtype(d["_ArrayType_"])
+                    )
                 if "_ArrayZipSize_" in d and newobj.shape[0] == 1:
                     if isinstance(d["_ArrayZipSize_"], str):
-                        d["_ArrayZipSize_"] = np.frombuffer(bytearray(d["_ArrayZipSize_"]))
+                        d["_ArrayZipSize_"] = np.frombuffer(
+                            bytearray(d["_ArrayZipSize_"])
+                        )
                     newobj = newobj.reshape(d["_ArrayZipSize_"])
                 if "_ArrayIsComplex_" in d and newobj.shape[0] == 2:
                     newobj = newobj[0] + 1j * newobj[1]
                 if "_ArrayOrder_" in d and (
-                    d["_ArrayOrder_"].lower() == "c" or d["_ArrayOrder_"].lower() == "col" or d["_ArrayOrder_"].lower() == "column"
+                    d["_ArrayOrder_"].lower() == "c"
+                    or d["_ArrayOrder_"].lower() == "col"
+                    or d["_ArrayOrder_"].lower() == "column"
                 ):
                     newobj = newobj.reshape(d["_ArraySize_"], order="F")
                 else:
@@ -308,6 +379,33 @@ def decode(d, opt={}):
                     "JData",
                     "one and only one of _ArrayData_ or _ArrayZipData_ is required",
                 )
+        elif "_DataLink_" in d:
+            if opt["maxlinklevel"] > 0 and "_DataLink_" in data:
+                if isinstance(data["_DataLink_"], str):
+                    datalink = data["_DataLink_"]
+                    if re.search("\:\$", datalink):
+                        ref = re.search(
+                            "^(?P<proto>[a-zA-Z]+://)*(?P<path>.+)(?P<delim>\:)()*(?P<jsonpath>(?<=:)\$\d*\.*.*)*",
+                            datalink,
+                        )
+                    else:
+                        ref = re.search(
+                            "^(?P<proto>[a-zA-Z]+://)*(?P<path>.+)(?P<delim>\:)*(?P<jsonpath>(?<=:)\$\d*\..*)*",
+                            datalink,
+                        )
+                    if ref and ref.group("path"):
+                        uripath = ref.group("proto") + ref.group("path")
+                        newobj, fname = jdlink(uripath)
+                        if os.path.exists(fname):
+                            opt["maxlinklevel"] = opt["maxlinklevel"] - 1
+                            if ref.group("jsonpath"):
+                                newobj = jsonpath(newdata, ref.group("jsonpath"))
+                        return nrewobj
+                    else:
+                        raise Exception(
+                            "JData",
+                            "_DataLink_ contains invalid URL",
+                        )
         return decodedict(d, opt)
     else:
         return copy.deepcopy(d) if opt["inplace"] else d
@@ -384,3 +482,126 @@ def decodelist(d0, opt={}):
 
 
 # -------------------------------------------------------------------------------------
+
+
+def zlibencode(buf):
+    return zlib.compress(buf)
+
+
+# -------------------------------------------------------------------------------------
+
+
+def gzipencode(buf):
+    gzipper = zlib.compressobj(wbits=(zlib.MAX_WBITS | 16))
+    newbuf = gzipper.compress(buf)
+    newbuf += gzipper.flush()
+    return newbuf
+
+
+# -------------------------------------------------------------------------------------
+
+
+def lzmaencode(buf):
+    try:
+        try:
+            import lzma
+        except ImportError:
+            from backports import lzma
+    except Exception:
+        raise Exception(
+            "JData",
+            'you must install "lzma" module to compress with this format',
+        )
+    return lzma.compress(buf, lzma.FORMAT_ALONE)
+
+
+# -------------------------------------------------------------------------------------
+
+
+def lz4encode(buf):
+    try:
+        import lz4.frame
+    except ImportError:
+        raise Exception(
+            "JData",
+            'you must install "lz4" module to compress with this format',
+        )
+    return lz4.compress(buf.tobytes(), lzma.FORMAT_ALONE)
+
+
+# -------------------------------------------------------------------------------------
+
+
+def base64encode(buf):
+    return base64.b64encode(buf)
+
+
+# -------------------------------------------------------------------------------------
+
+
+def zlibdecode(buf):
+    return zlib.decompress(buf)
+
+
+# -------------------------------------------------------------------------------------
+
+
+def gzipdecode(buf):
+    return zlib.decompress(bytes(buf), zlib.MAX_WBITS | 32)
+
+
+# -------------------------------------------------------------------------------------
+
+
+def lzmadecode(buf):
+    try:
+        try:
+            import lzma
+        except ImportError:
+            from backports import lzma
+    except Exception:
+        raise Exception(
+            "JData",
+            'you must install "lzma" module to compress with this format',
+        )
+    newbuf = bytearray(buf)  # set length to -1 (unknown) if EOF appears
+    newbuf[5:13] = b"\xff\xff\xff\xff\xff\xff\xff\xff"
+    return lzma.decompress(newbuf, lzma.FORMAT_ALONE)
+
+
+# -------------------------------------------------------------------------------------
+
+
+def lz4decode(buf):
+    try:
+        import lz4.frame
+    except ImportError:
+        raise Exception(
+            "JData",
+            'you must install "lz4" module to compress with this format',
+        )
+    return lz4.frame.decompress(bytes(buf))
+
+
+# -------------------------------------------------------------------------------------
+
+
+def base64decode(buf):
+    return base64.b64decode(buf)
+
+
+# -------------------------------------------------------------------------------------
+
+
+def jdataencode(obj, **kwargs):
+    return encode(obj, **kwargs)
+
+
+# -------------------------------------------------------------------------------------
+
+
+def jdatadecode(obj, **kwargs):
+    return decode(obj, **kwargs)
+
+
+# -------------------------------------------------------------------------------------
